from __future__ import annotations

from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class Element(Protocol):
    def __init__(
        self,
        native_element: Any,
        parent: Element | None = None,
    ) -> None:
        """
        Remote browser object API.

        Args:
            native_element (Any): The native element API for the underlying
                automation suite.
            parent (Element | None, optional): The parent. Defaults to None.
        """
        ...

    @property
    def parent(self) -> Element | None:
        """The parent element this was selected from, if any."""
        ...

    def query_selector_all(self, selector: str) -> list[Element] | None:
        """
        Select all child elements of this node that match the given selector.

        Args:
            selector (str): A CSS or XPATH selector string.

        Returns:
            list[Element] | None: A list of elements found, if any.
                Otherwise, None.
        """
        ...

    def query_selector(self, selector: str) -> Element | None:
        """
        Select the first child element of this node that matches the given
        selector.

        Args:
            selector (str): A CSS or XPATH selector string.

        Returns:
            list[Element] | None: Found element, if any. Otherwise, None.
        """
        ...

    def click(self) -> None:
        """Click on this element."""
        ...

    def content(self) -> str:
        """Return the text content of this element."""
        ...

    def type_on(self, text: str, delay: float = 0) -> None:
        """
        Emulate typing on this element.

        Args:
            text (str): Text to type.
            delay (float, optional): Time in seconds to wait between keys.
                Defaults to 0.
        """
        ...



from typing import Literal, Protocol, runtime_checkable

from mahler.protocols.element import Element
from mahler.fingerprint import Fingerprint


@runtime_checkable
class Window(Protocol):
    def __init__(
        self,
        model: Literal["chrome", "firefox"],
        headless: bool = True,
        enable_javascript: bool = True,
        fingerprint: Fingerprint | None = None,
    ):
        """
        Remote browser interactive API.

        Args:
            model (Literal["chrome", "firefox"]): Browser type.
            headless (bool, optional): Run in headless mode. Defaults to True.
            enable_javascript (bool, optional): Enable JavaScript in browser.
                Defaults to True.
            fingerprint (Fingerprint | None, optional): Browser fingerprint
                to apply, if any. Defaults to None.
        """
        ...

    def goto(self, url: str, timeout: float | None = None) -> None:
        """
        Navigate to a URL.

        Args:
            url (str): URL to visit.
            timeout (float | None, optional): Time in seconds to let load.
                Defaults to None.
        """
        ...

    def query_selector_all(self, selector: str) -> list[Element] | None:
        """
        Select all elements on the page that match the given selector.

        Args:
            selector (str): A CSS or XPATH selector string.

        Returns:
            list[Element] | None: A list of elements found, if any.
                Otherwise, None.
        """
        ...

    def query_selector(self, selector: str) -> Element | None:
        """
        Select the first element on the page that matches the given selector.

        Args:
            selector (str): A CSS or XPATH selector string.

        Returns:
            list[Element] | None: Found element, if any. Otherwise, None.
        """
        ...



import subprocess
from typing import Literal

from playwright.sync_api import sync_playwright, Page, BrowserType

from mahler.engines.playwright_.element import PlaywrightElement
from mahler.fingerprint import Fingerprint


def _install_playwright() -> None:
    """Install Playwright browsers via subprocess."""
    try:
        subprocess.check_output(
            ["playwright", "install", "--with-deps"],
        )
    except subprocess.CalledProcessError:
        raise RuntimeError("Failed to find or install browsers.")


def _create_browser(
    model: Literal["chrome", "firefox"],
    headless: bool = True,
    enable_javascript: bool = True,
    user_agent: str | None = None,
) -> Page:
    """Create Playwright browser, context, and page."""
    if model == "chrome":
        model = "chromium"
    playwright = sync_playwright().start()
    launcher: BrowserType = getattr(playwright, model)
    browser = launcher.launch(headless=headless)
    context = browser.new_context(
        java_script_enabled=enable_javascript,
        user_agent=user_agent,
    )
    page = context.new_page()
    return page


class PlaywrightWindow:
    """
    Playwright implementation of a Window.

    See mahler.protocols.window.Window for API details.
    """
    def __init__(
        self,
        model: Literal["chrome", "firefox"],
        headless: bool = True,
        enable_javascript: bool = True,
        fingerprint: Fingerprint | None = None,
    ):
        _install_playwright()
        self._page = _create_browser(
            model=model,
            headless=headless,
            enable_javascript=enable_javascript,
            user_agent=fingerprint.user_agent if fingerprint else None,
        )

    def goto(self, url: str, timeout: float | None = None) -> None:
        if timeout:
            timeout = timeout * 1000
        self._page.goto(url, wait_until="load", timeout=timeout)

    def query_selector_all(
        self,
        selector: str,
    ) -> list[PlaywrightElement] | None:
        element_handles = self._page.query_selector_all(selector)
        if not element_handles:
            return None
        return [PlaywrightElement(e) for e in element_handles]

    def query_selector(self, selector: str) -> PlaywrightElement | None:
        element_handle = self._page.query_selector(selector)
        if not element_handle:
            return None
        return PlaywrightElement(element_handle)



from __future__ import annotations

import time

from selenium.common.exceptions import (
    InvalidSelectorException,
    NoSuchElementException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement


def _find_elements(
    finder: WebDriver | WebElement,
    selector: str,
    first: bool,
) -> None | WebElement | list[WebElement]:
    """Find elements on a given Selenium driver or element."""
    method = finder.find_element if first else finder.find_elements
    try:
        web_elements = method(By.XPATH, selector)
        # Always try with CSS if xpath fails.
        if not web_elements:
            raise NoSuchElementException
    except (NoSuchElementException, InvalidSelectorException):
        try:
            web_elements = method(By.CSS_SELECTOR, selector)
        except (NoSuchElementException, InvalidSelectorException):
            return None
    return web_elements


class SeleniumElement:
    """
    Selenium implementation of an Element.

    See mahler.protocols.element.Element for API details.
    """
    def __init__(
        self,
        web_element: WebElement,
        parent: SeleniumElement | None = None,
    ) -> None:
        self._web_element = web_element
        self._parent = parent

    def __repr__(self):
        return f"Element <SeleniumElement {id(self)}>"

    @property
    def parent(self) -> SeleniumElement | None:
        return self._parent

    def query_selector_all(self, selector: str) -> list[SeleniumElement] | None:
        web_elements = _find_elements(self._web_element, selector, False)
        if not web_elements:
            return None
        return [SeleniumElement(e,  self) for e in web_elements]

    def query_selector(self, selector: str) -> SeleniumElement | None:
        web_element = _find_elements(self._web_element, selector, True)
        if not web_element:
            return None
        return SeleniumElement(web_element, self)

    def click(self) -> None:
        self._web_element.click()

    def content(self) -> str:
        return self._web_element.text

    def type_on(self, text: str, delay: float = 0) -> None:
        if not delay:
            self._web_element.send_keys(text)
        else:
            for char in text:
                self._web_element.send_keys(char)
                time.sleep(delay)



from mahler.engines.playwright_.element import PlaywrightElement  # noqa
from mahler.engines.playwright_.window import PlaywrightWindow  # noqa
from mahler.engines.selenium_.element import SeleniumElement  # noqa
from mahler.engines.selenium_.window import SeleniumWindow  # noqa



import warnings
from typing import Literal

from mahler.engines import (PlaywrightWindow, SeleniumWindow)
from mahler.fingerprint import Fingerprint
from mahler.protocols import Element, Window


ENGINE_TO_WINDOW_CLS = {
    "playwright": PlaywrightWindow,
    "selenium": SeleniumWindow,
}


class Controller:
    def __init__(
        self,
        engine: Literal["playwright", "selenium"],
        model: Literal["chrome", "firefox"],
        headless: bool = True,
        enable_javascript: bool = True,
        fingerprint: Fingerprint | None = None,
        user_agent: str | None = None,
    ):
        """
        Controller class to interact with a browser window.

        Args:
            engine (Literal["playwright", "selenium"]): Automation suite
                to use.
            model (Literal["chrome", "firefox"]): Browser type.
            headless (bool, optional): Run in headless mode. Defaults to True.
            enable_javascript (bool, optional): Enable JavaScript in browser.
                Defaults to True.
            fingerprint (Fingerprint | None, optional): Browser fingerprint
                to apply, if any. Defaults to None.
            user_agent (str | None, optional): User agent to apply to browser,
                if any. Defaults to None.

        Raises:
            ValueError: Raised if an invalid engine is given.
        """
        self._engine = engine
        self._model = model
        if user_agent:
            if fingerprint:
                warnings.warn("User agent ignored, fingerprint given.")
            else:
                fingerprint = Fingerprint(headers=None, user_agent=user_agent)
        window_cls: Window = ENGINE_TO_WINDOW_CLS.get(engine)
        if not window_cls:
            raise ValueError("Invalid engine given.")
        self._window: Window = window_cls(
            model,
            headless=headless,
            enable_javascript=enable_javascript,
            fingerprint=fingerprint,
        )

    def __repr__(self) -> str:
        kind = f"{self.engine.title()} {self.model.title()}"
        return f"Controller <{kind} at {id(self)}>"

    @property
    def engine(self) -> str:
        """Automation suite being used."""
        return self._engine

    @property
    def model(self) -> str:
        """Browser type."""
        return self._model

    @property
    def window(self) -> Window:
        """Current window object for the controller."""
        return self._window

    def goto(self, url: str, timeout: float | None = None) -> None:
        """
        Navigate to a URL.

        Args:
            url (str): URL to visit.
            timeout (float | None, optional): Time in seconds to let load.
                Defaults to None.
        """
        self._window.goto(url, timeout)

    def query_selector_all(self, selector: str) -> list[Element] | None:
        """
        Select all elements on the page that match the given selector.

        Args:
            selector (str): A CSS or XPATH selector string.

        Returns:
            list[Element] | None: A list of elements found, if any.
                Otherwise, None.
        """
        return self._window.query_selector_all(selector)

    def query_selector(self, selector: str) -> Element | None:
        """
        Select the first element on the page that matches the given selector.

        Args:
            selector (str): A CSS or XPATH selector string.

        Returns:
            list[Element] | None: Found element, if any. Otherwise, None.
        """
        return self._window.query_selector(selector)



from typing import Literal

from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.firefox.service import Service as FirefoxService
from selenium.webdriver.remote.webdriver import WebDriver
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager


from mahler.engines.selenium_.element import SeleniumElement, _find_elements
from mahler.fingerprint import Fingerprint


def _create_chromium_browser(
    headless: bool = True,
    enable_javascript: bool = True,
    user_agent: str | None = None,
) -> WebDriver:
    options = ChromeOptions()
    options.headless = headless
    if headless:
        options.add_argument('--headless=new')
    if not enable_javascript:
        prefs = {}
        prefs["webkit.webprefs.javascript_enabled"] = False
        prefs["profile.content_settings.exceptions.javascript.*.setting"] = 2
        prefs["profile.default_content_setting_values.javascript"] = 2
        prefs["profile.managed_default_content_settings.javascript"] = 2
        options.add_experimental_option("prefs", prefs)
        options.add_argument('--disable-javascript')
    if user_agent:
        options.add_argument(f"user-agent={user_agent}")
    return webdriver.Chrome(
        options=options,
        service=ChromeService(ChromeDriverManager().install())
    )


def _create_firefox_browser(
    headless: bool = True,
    enable_javascript: bool = True,
    user_agent: str | None = None,
) -> WebDriver:
    options = FirefoxOptions()
    options.headless = headless
    options.set_preference("javascript.enabled", enable_javascript)
    if user_agent:
        options.set_preference("general.useragent.override", user_agent)
    return webdriver.Firefox(
        options=options,
        service=FirefoxService(GeckoDriverManager().install())
    )


def _create_browser(
    model: Literal["chrome", "firefox"],
    headless: bool = True,
    enable_javascript: bool = True,
    user_agent: str | None = None,
) -> WebDriver:
    match model:
        case "chrome":
            return _create_chromium_browser(
                headless=headless,
                enable_javascript=enable_javascript,
                user_agent=user_agent,
            )
        case "firefox":
            return _create_firefox_browser(
                headless=headless,
                enable_javascript=enable_javascript,
                user_agent=user_agent,
            )
    raise RuntimeError("Invalid model given, can't create browser.")


class SeleniumWindow:
    """
    Selenium implementation of a Window.

    See mahler.protocols.window.Window for API details.
    """
    def __init__(
        self,
        model: Literal["chrome", "firefox"],
        headless: bool = True,
        enable_javascript: bool = True,
        fingerprint: Fingerprint | None = None,
    ):
        self._driver = _create_browser(
            model=model,
            headless=headless,
            enable_javascript=enable_javascript,
            user_agent=fingerprint.user_agent if fingerprint else None,
        )

    def goto(self, url: str, timeout: float | None = None) -> None:
        if timeout:
            self._driver.set_page_load_timeout(timeout)
        self._driver.get(url)

    def query_selector_all(self, selector: str) -> list[SeleniumElement] | None:
        web_elements = _find_elements(self._driver, selector, False)
        if not web_elements:
            return None
        return [SeleniumElement(e) for e in web_elements]

    def query_selector(self, selector: str) -> SeleniumElement | None:
        web_element = _find_elements(self._driver, selector, True)
        if not web_element:
            return None
        return SeleniumElement(web_element)



from __future__ import annotations

from playwright.sync_api import ElementHandle


class PlaywrightElement:
    """
    Playwright implementation of an Element.

    See mahler.protocols.element.Element for API details.
    """
    def __init__(
        self,
        element_handle: ElementHandle,
        parent: PlaywrightElement | None = None,
    ) -> None:
        self._element_handle = element_handle
        self._parent = parent

    def __repr__(self):
        return f"Element <PlaywrightElement {id(self)}>"

    @property
    def parent(self) -> PlaywrightElement | None:
        return self._parent

    def query_selector_all(
        self,
        selector: str,
    ) -> list[PlaywrightElement] | None:
        element_handles = self._element_handle.query_selector_all(selector)
        if not element_handles:
            return None
        return [PlaywrightElement(e,  self) for e in element_handles]

    def query_selector(self, selector: str) -> PlaywrightElement | None:
        element_handle = self._element_handle.query_selector(selector)
        if not element_handle:
            return None
        return PlaywrightElement(element_handle, self)

    def click(self) -> None:
        self._element_handle.click()

    def content(self) -> str:
        return self._element_handle.text_content()

    def type_on(self, text: str, delay: float = 0) -> None:
        delay = delay * 1000
        self._element_handle.type(text, delay=delay)
